/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Netscape security libraries.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corporation.
 * Portions created by the Initial Developer are Copyright (C) 1994-2000
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Hanno Boeck <hanno@hboeck.de>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
#include "cryptohi.h"
#include "secasn1.h"
#include "secitem.h"
#include "secder.h"
#include "secerr.h"
#include "pkcs11t.h"
#include "pk11pub.h"

/*
** Decode a DER encoded RSA-PSS parameter block to a PKCS #11 structure
** and get OID tag of hash algorithm. The latter can be NULL if it's
** not needed
*/
SECStatus
PSSU_DecodeDERPSSParams(CK_RSA_PKCS_PSS_PARAMS *dest, SECOidTag *hashAlg, 
               const SECItem *in)
{
    SECKEYRSAPSSParams pss_params;
    PRArenaPool *arena;
    SECStatus rv;
    SECAlgorithmID algId;
    SECOidTag oid, oid2;

    arena = PORT_NewArena(DER_DEFAULT_CHUNKSIZE);
    if (!arena) {
	PORT_SetError(SEC_ERROR_NO_MEMORY);
	return SECFailure;
    }
    PORT_Memset(&pss_params, 0, sizeof pss_params);
    rv = SEC_QuickDERDecodeItem(arena, &pss_params,
				SEC_ASN1_GET(SECKEY_RSAPSSParamsTemplate),
				in);
    if (rv != SECSuccess) {
	PORT_SetError(SEC_ERROR_INVALID_ARGS);
	goto finish;
    }

    /* Hash algorithm */
    if (pss_params.hashAlg) {
	oid = SECOID_FindOIDTag(&pss_params.hashAlg->algorithm);
    } else {
	oid = SEC_OID_SHA1;
    }
    if (hashAlg!=NULL) {
	*hashAlg = oid;
    }

    if (dest == NULL) {
	rv = SECSuccess;
	goto finish;
    }

    dest->hashAlg = PK11_AlgtagToMechanism(oid);

    /* According to RFC 4055, only five algorithms supported */
    if ((dest->hashAlg != CKM_SHA_1) && (dest->hashAlg != CKM_SHA256) &&
       (dest->hashAlg != CKM_SHA384) && (dest->hashAlg != CKM_SHA512) &&
       (dest->hashAlg != CKM_SHA224)) {
	PORT_SetError(SEC_ERROR_INVALID_ALGORITHM);
	rv = SECFailure;
	goto finish;
    }

    /* Mask Generation Function */
    if (pss_params.maskAlg) {
	rv = SEC_QuickDERDecodeItem(arena, &algId,
				    SEC_ASN1_GET(SECOID_AlgorithmIDTemplate),
				    &pss_params.maskAlg->parameters);
	if (rv != SECSuccess) {
	    PORT_FreeArena(arena, PR_FALSE);
	    rv = SECFailure;
	    goto finish;
	}

	oid = SECOID_FindOIDTag(&pss_params.maskAlg->algorithm);
	oid2 = SECOID_FindOIDTag(&algId.algorithm);
	dest->mgf = PK11_MgfMap(oid, oid2);
	if (dest->mgf == NULL) {
	    PORT_FreeArena(arena, PR_FALSE);
	    rv = SECFailure;
	    goto finish;
	}
    } else {
	dest->mgf = CKG_MGF1_SHA1;
    }

    /* Salt length */
    if (pss_params.saltLength.len > 0) {
	dest->sLen = DER_GetInteger(&pss_params.saltLength);
    } else {
	dest->sLen = 20;
    }

    /* Trailer field (must be absent or 1) */
    if (pss_params.trailerField.len > 0) {
	if (DER_GetInteger(&pss_params.trailerField) != 1) {
	    PORT_SetError(SEC_ERROR_INVALID_ARGS);
	    rv = SECFailure;
	    goto finish;
	}
    }

    rv = SECSuccess;

finish:
    PORT_FreeArena(arena, PR_FALSE);

    return rv;
}

SECStatus
PSSU_EncodeDERPSSParams(SECItem *dest, CK_RSA_PKCS_PSS_PARAMS *in)
{
    SECKEYRSAPSSParams *pss_params;
    PRArenaPool *arena;
    SECItem *ret;
    unsigned int i;
    SECOidTag hashAlg, maskAlg, maskHashAlg;
    SECStatus rv;
    SECItem *params;
    SECAlgorithmID algId;

    hashAlg = PK11_MechanismToAlgtag(in->hashAlg);

    if (in->mgf == CKG_MGF1_SHA1) {
	maskAlg = SEC_OID_PKCS1_MGF1;
	maskHashAlg = SEC_OID_SHA1;
    } else if (in->mgf == CKG_MGF1_SHA256) {
	maskAlg = SEC_OID_PKCS1_MGF1;
	maskHashAlg = SEC_OID_SHA256;
    } else if (in->mgf == CKG_MGF1_SHA384) {
	maskAlg = SEC_OID_PKCS1_MGF1;
	maskHashAlg = SEC_OID_SHA384;
    } else if (in->mgf == CKG_MGF1_SHA512) {
	maskAlg = SEC_OID_PKCS1_MGF1;
	maskHashAlg = SEC_OID_SHA512;
    } else {
	PORT_SetError(SEC_ERROR_INVALID_ARGS);
	return SECFailure;
    }

    arena = PORT_NewArena(DER_DEFAULT_CHUNKSIZE);
    pss_params = PORT_ArenaZNew(arena, SECKEYRSAPSSParams);

    if (hashAlg != SEC_OID_SHA1) {
	pss_params->hashAlg = PORT_ArenaZAlloc(arena, sizeof(SECAlgorithmID));
	SECOID_SetAlgorithmID(arena, pss_params->hashAlg, hashAlg, NULL);
    }

    if (maskHashAlg != SEC_OID_SHA1) {
	pss_params->maskAlg = PORT_ArenaZAlloc(arena, sizeof(SECAlgorithmID));
	bzero(&algId, sizeof(SECAlgorithmID));
	SECOID_SetAlgorithmID(arena, &algId, maskHashAlg, NULL);
	params = SEC_ASN1EncodeItem(arena, NULL, &algId,
				    SEC_ASN1_GET(SECOID_AlgorithmIDTemplate));
	SECOID_SetAlgorithmID(arena, pss_params->maskAlg, maskAlg, params);
    }

    if (in->sLen != 20) {
	DER_SetUInteger(arena, &pss_params->saltLength, in->sLen);
    }

    /* no non-standard trailerFields supported, see RFC 3447, A.2.3 */
    pss_params->trailerField.data = NULL;
    ret = SEC_ASN1EncodeItem(arena, NULL, pss_params,
			     SECKEY_RSAPSSParamsTemplate);

    dest->len=ret->len;
    dest->data=PORT_Alloc(dest->len);
    PORT_Memcpy(dest->data, ret->data, ret->len);

    PORT_FreeArena(arena, PR_FALSE);

    return SECSuccess;
}
